#![allow(non_snake_case)]
use argh::FromArgs;
use image::{self, imageops, Rgb};
use std::fs;
use std::io::{self, BufRead, BufReader, Read};
#[derive(FromArgs)]
/// Convert images to ascii.
/// works with Jpeg, PNG, and BMP formats
struct ASCIIfryer {
    /// width in columns of output ASCII
    #[argh(option, short = 'w', default = "40")]
    width: u32,
    /// height in rows of output ASCII
    #[argh(option, short = 'h', default = "40")]
    height: u32,
    /// palette of ASCII characters to render image with
    #[argh(option, short = 'p', default = "String::from(\".,-+=oO0#\")")]
    palette: String,

    /// read image from standard input
    #[argh(switch, short = 's')]
    stdin: bool,

    /// input file path (jpeg/png/bmp)
    #[argh(positional)]
    input: Option<String>,
}

struct Charset {
    characters: String,
    mult: f32,
}
impl Charset {
    fn new(characters: &str) -> Self {
        let characters = characters.to_owned();
        let mult = characters.chars().count() as f32 / (255.0 * 3.0);

        Charset { characters, mult }
    }
    fn convert_pixel(&self, px: &Rgb<u8>) -> char {
        let p = px[0] as u16 + px[1] as u16 + px[2] as u16;
        let scaled = p as f32 * self.mult;

        let scaled = scaled.floor() as usize;
        self.characters.chars().nth(scaled).unwrap_or('X')
    }
}

fn main() -> Result<(), &'static str> {
    // Args
    let settings: ASCIIfryer = argh::from_env();

    // Specifying whether reading from file or from stdin
    // invalid states panic
    let mut reader: Box<dyn BufRead> = match (settings.input, settings.stdin) {
        (Some(filename), false) => Box::new(BufReader::new(
            fs::File::open(filename).expect("Could not read file"),
        )),
        (None, true) => Box::new(BufReader::new(io::stdin().lock())),
        (Some(_), true) => {
            return Err("Can not load image from file and stdin. Use one or the other");
        }
        (None, false) => {
            return Err("No input provided");
        }
    };
    let mut file = vec![];
    // Load data from file/stdin
    reader
        .read_to_end(&mut file)
        .expect("Error while reading input");
    // Parse data to image
    let image = image::load_from_memory(&file)
        .expect("Could not load as image. PNG, BMP, and JPEG are supported");
    // Downscale image to reasonable size
    let image = image.resize(
        settings.width,
        settings.height,
        imageops::FilterType::Nearest,
    );
    // Convert image into buffer of Rgb<u8> (pixels)
    let buffer = image
        .as_rgb8()
        .expect("Could not extract 8-bit RGB buffer from image");

    let width = image.width();
    let charset = Charset::new(&settings.palette);

    // Iterate through each pixel and print resulting ASCII
    for (col, _, pix) in buffer.enumerate_pixels() {
        let symbol = charset.convert_pixel(pix);
        print!("{symbol}");
        if col == width - 1 {
            println!();
        }
    }
    Ok(())
}
